package intient.xml.bind;

import intient.xml.exception.MarshallerException;

import org.w3c.dom.Document;

/**
 * Interface to all supported marshaling operations.
 * 
 * @param <T>
 *            The type of object being marshaled (A JAXB generated class)
 * 
 * @author Bradley Beddoes
 */
public interface Marshaller<T> {
	/**
	 * Marshals content that does not require any signing in the final generated XML document. This marshaller will
	 * validate the supplied object against schema. Uses default UTF-8 encoding.
	 * 
	 * @param xmlObj
	 *            JAXB created implementation of some defined schema object
	 * @return A byte array representation of marshaled XML document if no errors occur.
	 * @throws MarshallerException
	 *             if an error occurs marshaling the document.
	 */
	public byte[] marshalUnsigned(T xmlObj) throws MarshallerException;

	/**
	 * Marshals content that does not require any signing in the final generated XML document and requires a certain
	 * type of charset encoding on this call. This marshaller will validate the supplied object against schema.
	 * 
	 * @param xmlObj
	 *            JAXB created implementation of some defined schema object
	 * @param encoding
	 *            The encoding to utilize in the create byte array
	 * @return A byte array representation of marshaled XML document if no errors occur.
	 * @throws MarshallerException
	 *             When error occurs marshaling the document.
	 */
	public byte[] marshalUnsigned(T xmlObj, String encoding) throws MarshallerException;

	/**
	 * Marshals content that requires enveloped XML signature creation. This marshaller will validate the supplied
	 * object against schema, the final document will include enveloped digital signatures for the elements listed Uses
	 * default UTF-8 encoding.
	 * 
	 * @param xmlObj
	 *            JAXB object of type to be marshaled, MUST contain 1..n empty <ds:Signature/> elements
	 * @return A byte array representation of marshaled XML document if no errors occur.
	 * @throws MarshallerException
	 *             When error occurs marshaling the document.
	 */
	public byte[] marshalSigned(T xmlObj) throws MarshallerException;

	/**
	 * Marshals content that requires enveloped XML signature creation and requires a certain type of charset encoding
	 * on this call This marshaller will validate the supplied object against schema, the final document will include
	 * enveloped digital signatures for the elements listed
	 * 
	 * @param xmlObj
	 *            JAXB object of type to be marshaled, MUST contain 1..n empty <ds:Signature/> elements
	 * @param encoding
	 *            The encoding to utilize in the create byte array
	 * @return A byte array representation of marshaled XML document if no errors occur.
	 * @throws MarshallerException
	 *             When error occurs marshaling the document.
	 */
	public byte[] marshalSigned(T xmlObj, String encoding) throws MarshallerException;

	/**
	 * Marshals content that does not require any signing in the final generated XML document. This marshaller will
	 * validate the supplied object against schema. Uses default UTF-8 encoding.
	 * 
	 * @param xmlObj
	 *            JAXB created implementation of some defined schema object
	 * @return A w3c Node representation of marshaled XML document if no errors occur
	 * @throws MarshallerException
	 *             if an error occurs marshaling the document.
	 */
	public Document marshalUnsignedDocument(T xmlObj) throws MarshallerException;

	/**
	 * Marshals content that does not require any signing in the final generated XML document and requires a certain
	 * type of charset encoding on this call. This marshaller will validate the supplied object against schema.
	 * 
	 * @param xmlObj
	 *            JAXB created implementation of some defined schema object
	 * @param encoding
	 *            The encoding to utilize in the create byte array
	 * @return A w3c Node representation of marshaled XML document if no errors occur.
	 * @throws MarshallerException
	 *             When error occurs marshaling the document.
	 */
	public Document marshalUnsignedDocument(T xmlObj, String encoding) throws MarshallerException;

	/**
	 * Marshals content that requires enveloped XML signature creation This marshaller will validate the supplied object
	 * against schema, the final document will include enveloped digital signatures for the elements listed Uses default
	 * UTF-8 encoding.
	 * 
	 * @param xmlObj
	 *            JAXB object of type to be marshaled, MUST contain 1..n empty <ds:Signature/> elements
	 * @return A w3c Node representation of marshaled XML document if no errors occur.
	 * @throws MarshallerException
	 *             When error occurs marshaling the document.
	 */
	public Document marshalSignedDocument(T xmlObj) throws MarshallerException;

	/**
	 * Marshals content that requires enveloped XML signature creation and requires a certain type of charset encoding
	 * on this call This marshaller will validate the supplied object against schema, the final document will include
	 * enveloped digital signatures for the elements listed
	 * 
	 * @param xmlObj
	 *            JAXB object of type to be marshaled, MUST contain 1..n empty <ds:Signature/> elements
	 * @param encoding
	 *            The encoding to utilize in the create byte array
	 * @return A w3c Node representation of marshaled XML document if no errors occur.
	 * @throws MarshallerException
	 *             When error occurs marshaling the document.
	 */
	public Document marshalSignedDocument(T xmlObj, String encoding) throws MarshallerException;
}
