package intient.xml.bind;

import intient.xml.exception.ReferenceValueException;
import intient.xml.exception.SignatureValueException;
import intient.xml.exception.UnmarshallerException;

import java.security.PublicKey;

import org.w3c.dom.Document;

/**
 * Interface to all supported unmarshalling operations
 * 
 * @param <T>
 *            The type of object being unmarshaled (A JAXB generated type)
 * 
 * @author Bradley Beddoes
 */
public interface Unmarshaller<T> {
	/**
	 * Validates document against schema, validates XML signature(s) and unmarshals supplied document to JAXB compiled
	 * object represented by T. Utilized where public key information is extracted from submitted document and
	 * implementation specific resolvers.
	 * 
	 * @param document
	 *            The byte array to unmarshal
	 * @return The JAXB unmarshaled object or null if an error occurs.
	 * @throws SignatureValueException
	 *             When the signature cannot be decoded.
	 * @throws ReferenceValueException
	 *             When the reference cannot be decoded.
	 * @throws UnmarshallerException
	 *             When an error occurs unmarshalling the document.
	 */
	public T unmarshalSigned(byte[] document) throws SignatureValueException, ReferenceValueException,
			UnmarshallerException;

	/**
	 * Validates document against schema, validates XML signature(s) and unmarshals supplied document to JAXB compiled
	 * object represented by T. Utilized where public key information is known in advance and supplied in call.
	 * 
	 * @param pk
	 *            The public key used to validate signature blocks
	 * @param document
	 *            The byte array to unmarshal
	 * @return The JAXB unmarshaled object or null if an error occurs.
	 * @throws SignatureValueException
	 *             When the signature cannot be decoded.
	 * @throws ReferenceValueException
	 *             When the reference cannot be decoded.
	 * @throws UnmarshallerException
	 *             When an error occurs unmarshalling the document.
	 */
	public T unmarshalSigned(PublicKey pk, byte[] document) throws SignatureValueException, ReferenceValueException,
			UnmarshallerException;

	/**
	 * Validates XML signature(s) and unmarshals supplied Node to JAXB compiled object represented by T. Utilized where
	 * public key information is extracted from submitted Node and implementation specific resolvers.
	 * 
	 * @param document
	 *            The node to unmarshal
	 * @return The JAXB unmarshaled object or null if an error occurs.
	 * @throws SignatureValueException
	 *             When the signature cannot be decoded.
	 * @throws ReferenceValueException
	 *             When the reference cannot be decoded.
	 * @throws UnmarshallerException
	 *             When an error occurs unmarshalling the document.
	 */
	public T unmarshalSigned(Document document) throws SignatureValueException, ReferenceValueException,
			UnmarshallerException;

	/**
	 * Validates XML signature(s) and unmarshals supplied Node to JAXB compiled object represented by T. Utilized where
	 * public key information is known in advance and supplied in call.
	 * 
	 * @param pk
	 *            The public key used to validate signature blocks
	 * @param document
	 *            The node to unmarshal
	 * @return The JAXB unmarshaled object or null if an error occurs.
	 * @throws SignatureValueException
	 *             When the signature cannot be decoded.
	 * @throws ReferenceValueException
	 *             When the reference cannot be decoded.
	 * @throws UnmarshallerException
	 *             When an error occurs unmarshalling the document.
	 */
	public T unmarshalSigned(PublicKey pk, Document document) throws SignatureValueException, ReferenceValueException,
			UnmarshallerException;

	/**
	 * Validates document against schema and unmarshals supplied document to JAXB compiled object represented by T
	 * 
	 * @param document
	 *            The byte array to unmarshal
	 * @return The JAXB unmarshaled object or null if an error occurs.
	 * @throws UnmarshallerException
	 *             When an error occurs unmarshalling the document.
	 */
	public T unmarshalUnsigned(byte[] document) throws UnmarshallerException;

	/**
	 * Validates node against schema and unmarshals supplied node to JAXB compiled object represented by T.
	 * 
	 * @param node
	 *            The node to unmarshall
	 * @return The JAXB unmarshaled object or null if an error occurs.
	 * @throws UnmarshallerException
	 *             When an error occurs unmarshalling the document.
	 */
	public T unmarshalUnsigned(Document document) throws UnmarshallerException;
}
