package intient.xml.bind.impl;

import intient.xml.constants.Constants;
import intient.xml.exception.MarshallerException;
import intient.xml.keys.KeyPairSelector;
import intient.xml.resolver.ResourceResolver;

import java.io.ByteArrayOutputStream;
import java.net.URL;
import java.security.PrivateKey;
import java.security.Provider;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Properties;

import javax.xml.XMLConstants;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.Marshaller;
import javax.xml.crypto.dsig.CanonicalizationMethod;
import javax.xml.crypto.dsig.DigestMethod;
import javax.xml.crypto.dsig.Reference;
import javax.xml.crypto.dsig.SignatureMethod;
import javax.xml.crypto.dsig.SignedInfo;
import javax.xml.crypto.dsig.Transform;
import javax.xml.crypto.dsig.XMLSignature;
import javax.xml.crypto.dsig.XMLSignatureFactory;
import javax.xml.crypto.dsig.dom.DOMSignContext;
import javax.xml.crypto.dsig.keyinfo.KeyInfo;
import javax.xml.crypto.dsig.keyinfo.KeyInfoFactory;
import javax.xml.crypto.dsig.keyinfo.KeyName;
import javax.xml.crypto.dsig.keyinfo.X509Data;
import javax.xml.crypto.dsig.keyinfo.X509IssuerSerial;
import javax.xml.crypto.dsig.spec.ExcC14NParameterSpec;
import javax.xml.crypto.dsig.spec.TransformParameterSpec;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.sun.xml.internal.bind.marshaller.NamespacePrefixMapper;

/**
 * Concrete implementation of all supported marshaling operations
 * 
 * @author Bradley Beddoes
 */
public class MarshallerImpl<T> implements intient.xml.bind.Marshaller<T> {
	private TransformerFactory transFac;
	private ResourceResolver resourceResolver;
	private JAXBContext jaxbContext;
	private SchemaFactory schemaFactory;
	private Schema schema;
	private NamespacePrefixMapper namespacePrefixMapper;
	private MarshallerValidationHandler validationHandler;

	private String keyPairName;
	private PrivateKey priv;
	private Certificate cert;
	private KeyPairSelector keyPairSelector;

	private Logger logger = LoggerFactory.getLogger(MarshallerImpl.class.getName());

	/**
	 * Constructor for MarshallerImpl.
	 * 
	 * @param packageName
	 *            MUST be the package where xmlObj was created in original JAXB compilation, it may also contain
	 *            additional colon separated package names where other class files required to marshall an object have
	 *            been placed by JAXB
	 * @param schemaList
	 *            List of URL locations for schemas to validate against
	 * @param keyPairSelector
	 *            Resolver for crypto details for this marshaler to utilize
	 * @param namespacePrefixMapper
	 *            Prefix mapping implementation to aide human consumption of XML
	 * @throws MarshallerException
	 *             When an error occurs creating the marshaller.
	 */
	public MarshallerImpl(String packageName, List<URL> schemaList, KeyPairSelector keyPairSelector,
			NamespacePrefixMapper namespacePrefixMapper) throws MarshallerException {

		try {
			this.resourceResolver = new ResourceResolver();
			this.namespacePrefixMapper = namespacePrefixMapper;
			this.jaxbContext = JAXBContext.newInstance(packageName);
			this.validationHandler = new MarshallerValidationHandler();

			ArrayList<Source> schemaSource = new ArrayList<Source>();
			for (URL location : schemaList) {
				schemaSource.add(new StreamSource(location.openStream()));
			}

			this.schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
			this.schemaFactory.setResourceResolver(this.resourceResolver);
			this.schema = this.schemaFactory.newSchema(schemaSource.toArray(new Source[0]));
			this.transFac = TransformerFactory.newInstance();

			if (keyPairSelector != null) {
				this.keyPairSelector = keyPairSelector;
				this.keyPairName = this.keyPairSelector.getKeyAlias();
				this.priv = this.keyPairSelector.getPrivateKey();
				this.cert = this.keyPairSelector.getCertificate();
			}
		} catch (Exception e) {
			this.logger.error(e.getLocalizedMessage());
			this.logger.debug(e.getLocalizedMessage(), e);
			throw new MarshallerException(e.getMessage(), e);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Marshaller#marshallSigned(java.lang.Object)
	 */
	public byte[] marshalSigned(T xmlObj) throws MarshallerException {
		return this.marshalSigned(xmlObj, Constants.DEFAULT_CHARSET);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Marshaller#marshallSigned(java.lang.Object, java.lang.String)
	 */
	public byte[] marshalSigned(T xmlObj, String encoding) throws MarshallerException {
		Document document = marshall(xmlObj, encoding);
		signNode(document);
		return generateByteArray(document, encoding);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Marshaller#marshallUnsigned(java.lang.Object)
	 */
	public byte[] marshalUnsigned(T xmlObj) throws MarshallerException {
		return this.marshalUnsigned(xmlObj, Constants.DEFAULT_CHARSET);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Marshaller#marshallUnsigned(java.lang.Object, java.lang.String)
	 */
	public byte[] marshalUnsigned(T xmlObj, String encoding) throws MarshallerException {
		Document document = marshall(xmlObj, encoding);
		return generateByteArray(document, encoding);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Marshaller#marshallUnsignedElement(java.lang.Object)
	 */
	public Document marshalUnsignedDocument(T xmlObj) throws MarshallerException {
		return this.marshalUnsignedDocument(xmlObj, Constants.DEFAULT_CHARSET);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Marshaller#marshallUnSignedNode(java.lang.Object, java.lang.String)
	 */
	public Document marshalUnsignedDocument(T xmlObj, String encoding) throws MarshallerException {
		Document document = marshall(xmlObj, encoding);
		return document;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Marshaller#marshallSignedElement(java.lang.Object)
	 */
	public Document marshalSignedDocument(T xmlObj) throws MarshallerException {
		return this.marshalSignedDocument(xmlObj, Constants.DEFAULT_CHARSET);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Marshaller#marshallSignedElement(java.lang.Object, java.lang.String)
	 */
	public Document marshalSignedDocument(T xmlObj, String encoding) throws MarshallerException {
		Document document = marshall(xmlObj, encoding);
		signNode(document);
		return document;
	}

	/**
	 * Takes the supplied JAXB object and marshals to W3C Document
	 * 
	 * @param xmlObj
	 *            JAXB object to marshal
	 * @param encoding
	 *            Encoding to use UTF-8 or UTF-16
	 * @return W3C document representation of xmlObj
	 * @throws MarshallerException
	 */
	private Document marshall(T xmlObj, String encoding) throws MarshallerException {
		if (xmlObj == null) {
			this.logger.error("Invalid XML object supplied for marshalling");
			throw new IllegalArgumentException("Invalid XML object supplied for marshalling");
		}
		try {
			// DBF is not thread safe
			DocumentBuilderFactory docBuildFac = DocumentBuilderFactory.newInstance();
			docBuildFac.setNamespaceAware(true);
			docBuildFac.setValidating(false);
			docBuildFac.setAttribute("http://apache.org/xml/features/dom/defer-node-expansion", Boolean.FALSE);

			DocumentBuilder docBuilder = docBuildFac.newDocumentBuilder();
			Document document = docBuilder.newDocument();

			Marshaller marshaller = jaxbContext.createMarshaller();
			marshaller.setProperty("jaxb.encoding", encoding);
			marshaller.setProperty("com.sun.xml.internal.bind.namespacePrefixMapper", this.namespacePrefixMapper);
			marshaller.setSchema(this.schema);
			marshaller.setEventHandler(this.validationHandler);

			marshaller.marshal(xmlObj, document);
			return document;
		} catch (Exception e) {
			this.logger.error(e.getLocalizedMessage());
			this.logger.debug(e.getLocalizedMessage(), e);
			throw new MarshallerException(e.getMessage(), e);
		}
	}

	/**
	 * Generates a byte array of characters of the supplied document
	 * 
	 * @param document
	 *            Document to translate to byte array
	 * @param encoding
	 *            Encoding to use in byte array chars UTF-8 or UTF-16
	 * @return Byte array representation of document
	 * @throws MarshallerException
	 */
	private byte[] generateByteArray(Document document, String encoding) throws MarshallerException {
		try {
			Properties properties = new Properties();
			properties.setProperty(OutputKeys.ENCODING, encoding);
			Transformer trans = this.transFac.newTransformer();
			trans.setOutputProperties(properties);

			ByteArrayOutputStream bytes = new ByteArrayOutputStream();
			StreamResult streamResult = new StreamResult(bytes);
			trans.transform(new DOMSource(document), streamResult);
			return bytes.toByteArray();
		} catch (Exception e) {
			this.logger.debug(e.getLocalizedMessage(), e);
			throw new MarshallerException(e.getMessage(), e);
		}
	}

	/**
	 * Signs all empty <ds:Signature/> elements in the supplied document.
	 * 
	 * @param document
	 *            Document whose Signature elements should be processed
	 * @throws MarshallerException
	 */
	private void signNode(Document document) throws MarshallerException {
		XMLSignatureFactory xmlSigFac;
		DigestMethod digestMethod;
		CanonicalizationMethod canocMeth;
		SignatureMethod sigMeth;

		ArrayList<Transform> transformList;
		SignedInfo signedInfo;
		DOMSignContext domSignContext;
		Element signatureParent = null;
		Element signatureElement = null;
		XMLSignature signature;
		Element postSignature;
		Reference ref;
		KeyInfoFactory factory;
		KeyInfo keyInfo;
		KeyName keyName;
		NodeList nodeList;
		String id;

		if (this.keyPairSelector == null) {
			this.logger.error("Key pair resolver was not specified at construction signed operations can't be completed");
			throw new IllegalArgumentException(
					"Key resolver was not specified at construction signed operations can't be completed");
		}

		try {
			// XMLSignatureFactory instances are not thread safe outside static functions so create in local scope
			xmlSigFac = XMLSignatureFactory.getInstance(Constants.DOM_FACTORY, (Provider) Class.forName(
					System.getProperty(Constants.JSR_MECHANISM, Constants.JSR_PROVIDER)).newInstance());
			digestMethod = xmlSigFac.newDigestMethod(DigestMethod.SHA1, null);
			canocMeth = xmlSigFac.newCanonicalizationMethod(CanonicalizationMethod.EXCLUSIVE,
					(ExcC14NParameterSpec) null);

			if (this.priv.getAlgorithm().equals(Constants.RSA_KEY)) {
				this.logger.debug("Marshaller is using RSA key for cryptography");
				sigMeth = xmlSigFac.newSignatureMethod(SignatureMethod.RSA_SHA1, null);
			} else {
				if (this.priv.getAlgorithm().equals(Constants.DSA_KEY)) {
					this.logger.debug("Marshaller is using DSA key for cryptography");
					sigMeth = xmlSigFac.newSignatureMethod(SignatureMethod.DSA_SHA1, null);
				} else {
					this.logger.error("Unsupported key type supplied, must be of RSA or DSA only");
					throw new MarshallerException("Unsupported key type supplied, must be of RSA or DSA only");
				}
			}

			// Locate all the empty Signature elements we wish to populate
			nodeList = document.getElementsByTagNameNS(XMLSignature.XMLNS, Constants.SIGNATURE_ELEMENT);
			if (nodeList.getLength() <= 0) {
				// Client did not instantiate a <Signature/> element template correctly
				this.logger.error("Attempt to sign document with no empty place holder signature elements");
				throw new MarshallerException("Attempt to sign document with no empty place holder signature elements");
			}

			// Sign elements in reverse so that outer calculations take into account inner signature blocks
			for (int i = nodeList.getLength() - 1; i >= 0; i--) {
				// Create new transform object for each element being signed
				transformList = new ArrayList<Transform>();
				TransformParameterSpec transformSpec = null;

				Transform exc14nTransform = xmlSigFac.newTransform(Constants.EXC14NTRANS, transformSpec);
				Transform envTransform = xmlSigFac.newTransform(Constants.ENVTRANS, transformSpec);

				transformList.add(envTransform);
				transformList.add(exc14nTransform);

				// Get the signature element and parent element for this enveloped signature
				signatureElement = (Element) nodeList.item(i);

				signatureParent = (Element) nodeList.item(i).getParentNode();
				id = signatureParent.getAttribute(Constants.ID_ATTRIBUTE);

				if (id == null || id.length() <= 0) {
					this.logger.warn("No ID attribute present in parent element of Signature element");
					throw new MarshallerException("No ID attribute present in parent element of Signature element");
				}
				this.logger.debug("Processing signature for element with XML ID " + id);

				ref = xmlSigFac.newReference("#" + id, digestMethod, transformList, null, null);
				signedInfo = xmlSigFac.newSignedInfo(canocMeth, sigMeth, Collections.singletonList(ref));

				factory = KeyInfoFactory.getInstance(Constants.DOM_FACTORY, (Provider) Class.forName(
						System.getProperty(Constants.JSR_MECHANISM, Constants.JSR_PROVIDER)).newInstance());
				keyName = factory.newKeyName(this.keyPairName);

				List<Object> keyInfoList = new ArrayList<Object>();
				keyInfoList.add(keyName);

				if (this.cert instanceof X509Certificate) {
					X509Certificate x509Certificate = (X509Certificate) this.cert;
					X509IssuerSerial x509IssuerSerial = factory.newX509IssuerSerial(x509Certificate.getIssuerDN()
							.getName(), x509Certificate.getSerialNumber());
					X509Data x509Data = factory.newX509Data(Collections.singletonList(x509IssuerSerial));
					keyInfoList.add(x509Data);
				}

				// Future additions to the KeyInfo block should be extended here
				keyInfo = factory.newKeyInfo(keyInfoList);

				// Skip text nodes here (handles human generated xml correctly)
				if ((signatureElement.getNextSibling() != null)
						&& (signatureElement.getNextSibling().getNodeType() == Node.TEXT_NODE)) {
					postSignature = (Element) signatureElement.getNextSibling().getNextSibling();
				} else {
					postSignature = (Element) signatureElement.getNextSibling();
				}

				// Strip away any additional textnodes that are located (handles human generated xml correctly)
				if ((signatureElement.getPreviousSibling() != null)
						&& (signatureElement.getPreviousSibling().getNodeType() == Node.TEXT_NODE)) {
					signatureParent.removeChild(signatureElement.getPreviousSibling());
				}

				signatureParent.removeChild(signatureElement);
				signature = xmlSigFac.newXMLSignature(signedInfo, keyInfo);

				// Determine signature context dependant on document contents
				if (postSignature != null)
					domSignContext = new DOMSignContext(this.priv, signatureParent, postSignature);
				else
					domSignContext = new DOMSignContext(this.priv, signatureParent);

				domSignContext.putNamespacePrefix(XMLSignature.XMLNS, "ds");
				signature.sign(domSignContext);
			}

		} catch (Exception e) {
			this.logger.debug(e.getLocalizedMessage(), e);
			throw new MarshallerException(e.getMessage(), e);
		}
	}
}
