package intient.xml.bind.impl;

import intient.xml.constants.Constants;
import intient.xml.exception.MarshallerException;
import intient.xml.exception.ReferenceValueException;
import intient.xml.exception.SignatureValueException;
import intient.xml.exception.UnmarshallerException;
import intient.xml.keys.PublicKeySelector;
import intient.xml.keys.KeySelector;
import intient.xml.resolver.ResourceResolver;

import java.io.ByteArrayInputStream;
import java.net.URL;
import java.security.Provider;
import java.security.PublicKey;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.xml.XMLConstants;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.Unmarshaller;
import javax.xml.crypto.MarshalException;
import javax.xml.crypto.dsig.Reference;
import javax.xml.crypto.dsig.XMLSignature;
import javax.xml.crypto.dsig.XMLSignatureException;
import javax.xml.crypto.dsig.XMLSignatureFactory;
import javax.xml.crypto.dsig.dom.DOMValidateContext;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

/**
 * Concrete implementation of all supported unmarshalling operations
 * 
 * @author Bradley Beddoes
 */
public class UnmarshallerImpl<T> implements intient.xml.bind.Unmarshaller<T> {

	private JAXBContext jaxbContext;
	private ResourceResolver resourceResolver;
	private PublicKeySelector publicKeySelector;
	private SchemaFactory schemaFactory;
	private Schema schema;

	private Logger logger = LoggerFactory.getLogger(UnmarshallerImpl.class.getName());

	/**
	 * Constructor for UnmarshallerImpl.
	 * 
	 * @param packageName
	 *            MUST be the package where xmlObj was created in original JAXB compilation, it may also contain
	 *            additional colon separated package names where other class files required to unmarshall an object have
	 *            been placed by JAXB
	 * @param schemaList
	 *            List of URL locations for schemas to validate against
	 * @param publicKeySelector
	 *            Used to select keys to validate XML signatures from externally managed source
	 * @throws UnmarshallerException
	 *             When an error occurs creating the marshaller.
	 */
	public UnmarshallerImpl(String packageName, List<URL> schemaList, PublicKeySelector publicKeySelector)
			throws UnmarshallerException {

		try {
			this.resourceResolver = new ResourceResolver();
			this.jaxbContext = JAXBContext.newInstance(packageName);

			ArrayList<Source> schemaSource = new ArrayList<Source>();
			for (URL location : schemaList) {
				schemaSource.add(new StreamSource(location.openStream()));
			}

			this.schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
			this.schemaFactory.setResourceResolver(this.resourceResolver);
			this.schema = this.schemaFactory.newSchema(schemaSource.toArray(new Source[0]));

			this.publicKeySelector = publicKeySelector;

		} catch (Exception e) {
			this.logger.error(e.getLocalizedMessage());
			this.logger.debug(e.getLocalizedMessage(), e);
			throw new UnmarshallerException(e.getMessage(), e, null);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Unmarshaller#unmarshalSigned(java.security.PublicKey, byte[])
	 */
	public T unmarshalSigned(PublicKey pk, byte[] document) throws SignatureValueException, ReferenceValueException,
			UnmarshallerException {
		Document doc = translateDocument(document);
		T jaxbObject = unmarshal(doc);
		validateSignature(doc, pk, jaxbObject);
		return jaxbObject;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Unmarshaller#unmarshalSigned(byte[])
	 */
	public T unmarshalSigned(byte[] document) throws SignatureValueException, ReferenceValueException,
			UnmarshallerException {
		Document doc = translateDocument(document);
		T jaxbObject = unmarshal(doc);
		validateSignature(doc, null, jaxbObject);
		return jaxbObject;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Unmarshaller#unmarshalSigned(org.w3c.dom.Document)
	 */
	public T unmarshalSigned(Document document) throws SignatureValueException, ReferenceValueException, UnmarshallerException {
		T jaxbObject = unmarshal(document);
		validateSignature(document, null, jaxbObject);
		return jaxbObject;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Unmarshaller#unmarshalSigned(java.security.PublicKey, org.w3c.dom.Document)
	 */
	public T unmarshalSigned(PublicKey pk, Document document) throws SignatureValueException, ReferenceValueException,
			UnmarshallerException {
		T jaxbObject = unmarshal(document);
		validateSignature(document, pk, jaxbObject);
		return jaxbObject;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Unmarshaller#unmarshalUnsigned(org.w3c.dom.Document)
	 */
	public T unmarshalUnsigned(Document document) throws UnmarshallerException {
		T jaxbObject = unmarshal(document);
		return jaxbObject;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see intient.xml.bind.Unmarshaller#unmarshalUnsigned(byte[])
	 */
	public T unmarshalUnsigned(byte[] document) throws UnmarshallerException {
		Document doc = translateDocument(document);
		T jaxbObject = unmarshal(doc);
		return jaxbObject;
	}

	/**
	 * Takes a byte array and translates to w3c document representation
	 * 
	 * @param document
	 *            xml in byte array format to translate to w3c Document
	 * @return A w3c document representation of supplied byte array
	 * @throws UnmarshallerException
	 *             When the byte array is invalid
	 */
	private Document translateDocument(byte[] document) throws UnmarshallerException {
		if (document == null) {
			this.logger.error("Invalid byte array supplied to translate operation");
			throw new IllegalArgumentException("Invalid byte array supplied to translate operation");
		}
		
		try {
			DocumentBuilderFactory docBuildFac = DocumentBuilderFactory.newInstance();
			docBuildFac.setNamespaceAware(true);
			docBuildFac.setValidating(false);
			docBuildFac.setAttribute("http://apache.org/xml/features/dom/defer-node-expansion", Boolean.FALSE);

			DocumentBuilder docBuilder = docBuildFac.newDocumentBuilder();

			ByteArrayInputStream bais = new ByteArrayInputStream(document);
			return docBuilder.parse(bais);
		} catch (Exception e) {
			this.logger.warn(e.getLocalizedMessage());
			this.logger.debug(e.getLocalizedMessage(), e);
			throw new UnmarshallerException(e.getMessage(), e);
		}
	}

	/**
	 * Takes the supplied W3C Document and unmarshals to JAXB instance
	 * 
	 * @param xmlObj
	 *            JAXB object to marshal
	 * @param encoding
	 *            Encoding to use UTF-8 or UTF-16
	 * @return W3C document representation of xmlObj
	 * @throws MarshallerException
	 */
	private T unmarshal(Document document) throws UnmarshallerException {
		if (document == null) {
			this.logger.error("Invalid document supplied to unmarshal operation");
			throw new IllegalArgumentException("Invalid document supplied to unmarshal operation");
		}

		T jaxbObject = null;
		try {
			Unmarshaller unmarshaller = this.jaxbContext.createUnmarshaller();
			unmarshaller.setSchema(this.schema);

			jaxbObject = (T) unmarshaller.unmarshal(document);
			return jaxbObject;
		} catch (Exception e) {
			this.logger.debug(e.getLocalizedMessage(), e);
			throw new UnmarshallerException(e.getLocalizedMessage(), e, jaxbObject);
		}
	}

	/**
	 * Validates signature blocks in supplied w3c document.
	 * 
	 * @param doc
	 *            The document to validate signatures in.
	 * @param pk
	 *            The public key to validate signatures against. If null and publicKeySelector is valid this will be
	 *            used to resolve key information on a per signature block basis.
	 * @param jaxbObject
	 *            The already unmarshalled JAXB instance of the supplied document
	 * @throws UnmarshallerException
	 *             When a general error state occurs
	 * @throws SignatureValueException
	 *             When the signature cannot be decoded.
	 * @throws ReferenceValueException
	 *             When the reference cannot be decoded.
	 */
	private void validateSignature(Document doc, PublicKey pk, T jaxbObject) throws UnmarshallerException,
			SignatureValueException, ReferenceValueException {

		boolean validSig = false;
		NodeList nodeList;
		KeySelector keySelector;
		DOMValidateContext valContext;
		XMLSignature signature;
		XMLSignatureFactory xmlSigFac;

		try {
			// XMLSignatureFactory instances are not thread safe outside static functions so create in local scope
			xmlSigFac = XMLSignatureFactory.getInstance(Constants.DOM_FACTORY, (Provider) Class.forName(
					System.getProperty(Constants.JSR_MECHANISM, Constants.JSR_PROVIDER)).newInstance());
		} catch (Exception e) {
			this.logger.debug(e.getLocalizedMessage(), e);
			throw new UnmarshallerException(e.getLocalizedMessage(), e, jaxbObject);
		}

		nodeList = doc.getElementsByTagNameNS(XMLSignature.XMLNS, Constants.SIGNATURE_ELEMENT);
		if (nodeList == null || nodeList.getLength() == 0) {
			this.logger.warn("No ds:Signature blocks were located in supplied document, failing signature validation");
			throw new UnmarshallerException("No ds:Signature blocks were located in supplied document, failing signature validation", null, jaxbObject);
		}

		for (int i = 0; i < nodeList.getLength(); i++) {
			if (pk != null) {
				valContext = new DOMValidateContext(pk, nodeList.item(i));
			} else {
				if (this.publicKeySelector == null) {
					this.logger
							.error("Public key selector not provided to unmarshaller construction, unable to validate signatures without supplied public key");
					throw new IllegalArgumentException(
							"Public key selector not provided to unmarshaller construction, unable to validate signatures without supplied public key");
				}

				keySelector = new KeySelector(this.publicKeySelector);
				valContext = new DOMValidateContext(keySelector, nodeList.item(i));
			}

			try {
				signature = xmlSigFac.unmarshalXMLSignature(valContext);

				validSig = signature.validate(valContext);

				if (!validSig) {
					this.logger.warn("Document signature was determined to be invalid");

					// Signature fault, determine why for caller
					if (!signature.getSignatureValue().validate(valContext)) {
						this.logger.warn("Signature validation failed with invalid signature for supplied public key");
						throw new SignatureValueException("Signature validation failed with invalid signature for supplied public key", null, jaxbObject);
					}

					Iterator<Reference> j = signature.getSignedInfo().getReferences().iterator();
					while (j.hasNext()) {
						if (!j.next().validate(valContext)) {
							this.logger.warn("Signature validation failed with invalid supplied reference, the request document has been tampered with or the parent element to this signature block has no ID attribute");
							throw new ReferenceValueException("Signature validation failed with invalid supplied reference, the request document has been tampered with or the parent element to this signature block has no ID attribute", null,
									jaxbObject);
						}
					}

					// Can't ascertain what's gone wrong at this point
					this.logger.warn("Signature block is invalid but unable to determine specific reasoning");
					throw new UnmarshallerException("Signature block is invalid but unable to determine specific reasoning", null, jaxbObject);
				}
			} catch (XMLSignatureException e) {
				this.logger.debug(e.getLocalizedMessage(), e);
				throw new UnmarshallerException(e.getLocalizedMessage(), e, jaxbObject);
			} catch (MarshalException e) {
				this.logger.debug(e.getLocalizedMessage(), e);
				throw new UnmarshallerException(e.getLocalizedMessage(), e, jaxbObject);
			}
		}
	}
}
