package intient.xml.keys;

import intient.xml.exception.KeySelectorException;

import java.math.BigInteger;
import java.security.PublicKey;
import java.util.List;

import javax.xml.crypto.AlgorithmMethod;
import javax.xml.crypto.KeySelectorResult;
import javax.xml.crypto.XMLCryptoContext;
import javax.xml.crypto.XMLStructure;
import javax.xml.crypto.dsig.keyinfo.KeyInfo;
import javax.xml.crypto.dsig.keyinfo.KeyName;
import javax.xml.crypto.dsig.keyinfo.X509Data;
import javax.xml.crypto.dsig.keyinfo.X509IssuerSerial;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Selects keys by name from a configured external key selector
 * 
 * @author Bradley Beddoes
 */
public class KeySelector extends javax.xml.crypto.KeySelector {

	private PublicKeySelector externalKeySelector;

	private Logger logger = LoggerFactory.getLogger(KeySelector.class.getName());

	public KeySelector(PublicKeySelector externalKeySelector) {
		super();

		if (externalKeySelector == null) {
			this.logger.error("Configured external key selector is invalid");
			throw new IllegalArgumentException("Configured external key selector is invalid");
		}

		this.externalKeySelector = externalKeySelector;
	}

	public KeySelectorResult select(KeyInfo keyInfo, KeySelector.Purpose purpose, AlgorithmMethod method,
			XMLCryptoContext context) throws javax.xml.crypto.KeySelectorException {
		if (keyInfo == null) {
			this.logger.warn("KeyInfo is null unable to select key from null input");
			throw new javax.xml.crypto.KeySelectorException("KeyInfo is null unable to select key from null input");
		}

		String name = null;
		String issuerName = null;
		BigInteger serialNumber = null;
		ExternalKeySelectorResult result;
		List<XMLStructure> keyElementList = keyInfo.getContent();

		for (XMLStructure keyElement : keyElementList) {
			if (keyElement instanceof KeyName) {
				KeyName keyName = (KeyName) keyElement;
				name = keyName.getName();
				this.logger.debug("Located value for keyName of " + keyName.getName());
			} else if (keyElement instanceof X509Data) {
				X509Data x509Data = (X509Data) keyElement;
				for (Object x509DataElement : x509Data.getContent()) {
					if (x509DataElement instanceof X509IssuerSerial) {
						X509IssuerSerial x509IssuerSerial = (X509IssuerSerial) x509DataElement;
						issuerName = x509IssuerSerial.getIssuerName();
						serialNumber = x509IssuerSerial.getSerialNumber();
					}
				}
			}
		}

		if (issuerName != null && serialNumber != null) {
			this.logger
					.debug("KeyInfo had X509Data with issuer DN and serial. Using this as preferred method of resolving cert");
			this.logger.debug("X509Data had issuer DN '" + issuerName + "' and serial '" + serialNumber.toString()
					+ "'");
			try {
				PublicKey key = this.externalKeySelector.selectKey(issuerName, serialNumber);
				if (key != null) {
					result = new ExternalKeySelectorResult(key);
					this.logger
							.debug("Resolved certificate data successfully from X509Data issuer DN and serial number.");
					return result;
				}
			} catch (KeySelectorException e) {
				this.logger
						.warn("Could not resolve certificate from X509Data issuer DN and serial number. Trying to resolve by name");
			}
		}

		try {
			if (name == null) {
				throw new javax.xml.crypto.KeySelectorException(
						"KeyInfo block was not suitable for use with KeySelector, issuerName, serialNumber and Name are all null");
			}
			result = new ExternalKeySelectorResult(this.externalKeySelector.selectKey(name));
		} catch (Exception e) {
			this.logger.warn(e.getLocalizedMessage(), e);
			throw new javax.xml.crypto.KeySelectorException(e.getLocalizedMessage(), e);
		}

		return result;
	}
}
