package intient.xml.resolver;

import intient.xml.exception.ResourceException;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.ls.LSInput;
import org.w3c.dom.ls.LSResourceResolver;

/**
 * Resolves schema resources locally for common xml schema.
 * 
 * @author Bradley Beddoes
 */
public class ResourceResolver implements LSResourceResolver {
	private final String XMLDSIG = "xmldsig-core-schema.xsd";
	private final String XMLENC = "xenc-schema.xsd";
	private final String XML = "xml.xsd";

	private Logger logger = LoggerFactory.getLogger(ResourceResolver.class.getName());

	private Map<String, LSInput> schemas;

	public ResourceResolver() throws ResourceException {
		this.schemas = new HashMap<String, LSInput>();

		loadSchema(this.XMLDSIG);
		loadSchema(this.XMLENC);
		loadSchema(this.XML);
	}

	/**
	 * @param resource
	 *            The name of the schema file to load including extension
	 */
	private void loadSchema(String resource) throws ResourceException {
		String str;
		StringBuffer xml;
		Reader reader = null;
		BufferedReader in = null;
		InputStream fileStream = null;

		try {
			URL location = ResourceResolver.class.getResource(resource);
			LSInput input = new LSInputImpl();
			if (location == null)
				throw new IllegalArgumentException("Unable to locate resource " + resource
						+ " in ResourceResolver classpath");
			input.setBaseURI(location.getFile());

			xml = new StringBuffer();
			fileStream = location.openStream();
			reader = new InputStreamReader(fileStream);
			in = new BufferedReader(reader);

			while ((str = in.readLine()) != null) {
				xml.append(str);
			}

			input.setStringData(xml.toString());

			this.schemas.put(resource, input);
		} catch (IOException e) {
			this.logger.error("Exception when loading resource " + resource + " from ResourceResolver classpath");
			throw new ResourceException(e.getLocalizedMessage(), e);
		} finally {
			try {
				if (in != null)
					in.close();

				if (reader != null)
					reader.close();

				if (fileStream != null)
					fileStream.close();
			} catch (IOException e) {
				this.logger.error("Unable to close file streams when loading schema");
				this.logger.error(e.toString());
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.w3c.dom.ls.LSResourceResolver#resolveResource(java.lang.String, java.lang.String, java.lang.String,
	 * java.lang.String, java.lang.String)
	 */
	public LSInput resolveResource(String type, String namespaceURI, String publicId, String systemId, String baseURI) {
		if (namespaceURI != null && namespaceURI.equals("http://www.w3.org/2000/09/xmldsig#")) {
			return this.schemas.get(this.XMLDSIG);
		}

		if (namespaceURI != null && namespaceURI.equals("http://www.w3.org/2001/04/xmlenc#")) {
			return this.schemas.get(this.XMLENC);
		}

		if (namespaceURI != null && namespaceURI.equals("http://www.w3.org/XML/1998/namespace")) {
			return this.schemas.get(this.XML);
		}

		// Resolver doesn't know about this schema so pass back default object
		LSInput input = new LSInputImpl();
		input.setSystemId(systemId);
		input.setPublicId(publicId);
		input.setBaseURI(baseURI);
		return input;
	}
}
