package intient.xml.bind.impl

import groovy.util.GroovyTestCase
import groovy.mock.interceptor.MockFor

import org.oasis.saml2.assertion.*;
import org.oasis.saml2.protocol.*;
import org.w3._2000._09.xmldsig_.*;

import java.security.KeyStore
import java.security.PrivateKey
import java.security.cert.Certificate

import javax.xml.datatype.XMLGregorianCalendar;
import com.sun.org.apache.xerces.internal.jaxp.datatype.XMLGregorianCalendarImpl;
import com.sun.xml.internal.bind.marshaller.NamespacePrefixMapper;

import intient.xml.bind.impl.*
import intient.xml.constants.*
import intient.xml.keys.*
import intient.xml.namespace.*
import intient.xml.exception.*

class MarshallerImplTest extends GroovyTestCase {

	def keyAlias = 'testing'
	def packageName = "org.oasis.saml2.protocol"

	void testMarshalSigned() {
		def marshaller = setupMarshaller()
		def authnRequest = createAuthnRequest()

        def requestDocument = new String(marshaller.marshalSigned(authnRequest))
		assertTrue requestDocument.contains("AuthnRequest")
		assertTrue requestDocument.contains("<ds:SignatureValue>")
		assertTrue requestDocument.contains("AssertionConsumerServiceURL=\"http://service.example.com/sso/saml.post\"")
		assertTrue requestDocument.contains("ID=\"abe567de6-122wert67\"")
		assertTrue requestDocument.contains("NameID Format=\"urn:oasis:names:tc:SAML:2.0:nameid-format:transient\">1234567")
		assertTrue requestDocument.contains("Audience>service.example.com")	
	}

	void testMarshalSignedNoSigElement() {
		def marshaller = setupMarshaller()
		def authnRequest = createAuthnRequest()
		authnRequest.signature = null

		def failure = shouldFail(MarshallerException.class) {
			marshaller.marshalSigned(authnRequest)
		}
		assert failure.contains("Attempt to sign document with no empty place holder signature elements")	
	}

	void testSchemaInvalidMarshalSigned() {
		def marshaller = setupMarshaller()
		def authnRequest = createSchemaInvalidAuthnRequest()

		shouldFail(MarshallerException.class) {
			marshaller.marshalSigned(authnRequest)
		}
	}

	void testMarshallerSignedNullXMLObj() {
		def marshaller = setupMarshaller()
		def failure = shouldFail(IllegalArgumentException.class) {
			marshaller.marshalSigned(null)
		}
		assert failure.contains("Invalid XML object supplied for marshalling")
	}

	void testMarshallerSignedNullNoCrypto() {
		def schemaList = loadSchema()
    	def npm = [:] as NamespacePrefixMapperImpl
    	
    	def marshaller = new MarshallerImpl<AuthnRequest>(packageName, schemaList, null, npm)
    	AuthnRequest authnRequest = createAuthnRequest()
		
		def failure = shouldFail(IllegalArgumentException.class) {
			marshaller.marshalSigned(authnRequest)
		}
		assert failure.contains("Key resolver was not specified at construction signed operations can't be completed")
	}

	void testMarshalUnsigned() {
		def marshaller = setupMarshaller()
		def authnRequest = createAuthnRequest()

        def requestDocument = new String(marshaller.marshalUnsigned(authnRequest))
		assertTrue requestDocument.contains("AuthnRequest")
		assertTrue requestDocument.contains("<ds:Signature/>")
		assertTrue requestDocument.contains("AssertionConsumerServiceURL=\"http://service.example.com/sso/saml.post\"")
		assertTrue requestDocument.contains("ID=\"abe567de6-122wert67\"")
		assertTrue requestDocument.contains("NameID Format=\"urn:oasis:names:tc:SAML:2.0:nameid-format:transient\">1234567")
		assertTrue requestDocument.contains("Audience>service.example.com")
	}

	void testSchemaInvalidMarshalUnsigned() {
		def marshaller = setupMarshaller()
		def authnRequest = createSchemaInvalidAuthnRequest()

		shouldFail(MarshallerException.class) {
			marshaller.marshalUnsigned(authnRequest)
		}
	}

	void testMarshallerUnsignedNullXMLObj() {
		def marshaller = setupMarshaller()
		def failure = shouldFail(IllegalArgumentException.class) {
			marshaller.marshalUnsigned(null)
		}
		assert failure.contains("Invalid XML object supplied for marshalling")
	}

	void testMarshalSignedDocument() {
		def marshaller = setupMarshaller()
		def authnRequest = createAuthnRequest()

        def requestDocument = marshaller.marshalSignedDocument(authnRequest)
        assertTrue requestDocument.documentElement.getAttribute("AssertionConsumerServiceURL").equals("http://service.example.com/sso/saml.post")
        assertTrue requestDocument.documentElement.getAttribute("ID").equals("abe567de6-122wert67")
        assertTrue requestDocument.documentElement.childNodes.item(0).localName.equals ("Signature")
        assertTrue requestDocument.documentElement.childNodes.item(0).childNodes.length == 3
        assertTrue requestDocument.documentElement.childNodes.item(1).localName.equals ("Subject")
        assertTrue requestDocument.documentElement.childNodes.item(2).localName.equals ("Conditions")
	}

	void testMarshallerSignedDocumentNullXMLObj() {
		def marshaller = setupMarshaller()
		def failure = shouldFail(IllegalArgumentException.class) {
			marshaller.marshalSignedDocument(null)
		}
		assert failure.contains("Invalid XML object supplied for marshalling")
	}

	void testMarshallerSignedDocumentNullNoCrypto() {
		def schemaList = loadSchema()
    	def npm = [:] as NamespacePrefixMapperImpl
    	
    	def marshaller = new MarshallerImpl<AuthnRequest>(packageName, schemaList, null, npm)
    	AuthnRequest authnRequest = createAuthnRequest()
		
		def failure = shouldFail(IllegalArgumentException.class) {
			marshaller.marshalSignedDocument(authnRequest)
		}
		assert failure.contains("Key resolver was not specified at construction signed operations can't be completed")
	}

	void testMarshalUnsignedDocument() {
		def marshaller = setupMarshaller()
		def authnRequest = createAuthnRequest()

        def requestDocument = marshaller.marshalUnsignedDocument(authnRequest)
        assertTrue requestDocument.documentElement.getAttribute("AssertionConsumerServiceURL").equals("http://service.example.com/sso/saml.post")
        assertTrue requestDocument.documentElement.getAttribute("ID").equals("abe567de6-122wert67")
        assertTrue requestDocument.documentElement.childNodes.item(0).localName.equals ("Signature")
        assertTrue requestDocument.documentElement.childNodes.item(0).childNodes.length == 0
        assertTrue requestDocument.documentElement.childNodes.item(1).localName.equals ("Subject")
        assertTrue requestDocument.documentElement.childNodes.item(2).localName.equals ("Conditions")
	}

	void testMarshallerUnsignedDocumentNullXMLObj() {
		def marshaller = setupMarshaller()
		def failure = shouldFail(IllegalArgumentException.class) {
			marshaller.marshalUnsignedDocument(null)
		}
		assert failure.contains("Invalid XML object supplied for marshalling")
	}

	// Helpers
    def setupMarshaller() {
    	def schemaList = loadSchema()
    	def lkr = loadLocalKeys()
    	def npm = [:] as NamespacePrefixMapperImpl
    	
    	def marshaller = new MarshallerImpl<AuthnRequest>(packageName, schemaList, lkr, npm)
    }

    def loadSchema() {
    	def assertion = new File("src/test/groovy/intient/xml/bind/saml-schema-assertion-2.0.xsd")
    	def metadata = new File("src/test/groovy/intient/xml/bind/saml-schema-metadata-2.0.xsd")
    	def protocol = new File("src/test/groovy/intient/xml/bind/saml-schema-protocol-2.0.xsd")

    	def schemas = [assertion.toURL(), protocol.toURL(), metadata.toURL()]
        return schemas
    }

    def loadLocalKeys() {
    	KeyStore ks = KeyStore.getInstance("JKS")
        FileInputStream fis = new FileInputStream("src/test/groovy/intient/xml/bind/keystore.ks")
        char[] passwd = ['t', 'e', 's', 't', 'i', 'n', 'g']
        ks.load(fis, passwd)

        PrivateKey priv = (PrivateKey) ks.getKey(keyAlias, passwd)
        Certificate cert = ks.getCertificate(keyAlias)

        def lkr = [getKeyAlias:{keyAlias}, getPrivateKey:{priv}, getCertificate:{cert}] as KeyPairSelector
    }

    def createAuthnRequest() {
		AudienceRestriction audienceRestriction = new AudienceRestriction()
        Conditions conditions = new Conditions()
        NameIDType nameID = new NameIDType()
        Subject subject = new Subject()
        Signature signature = new Signature()
        AuthnRequest authnRequest = new AuthnRequest()

        SimpleTimeZone gmt = new SimpleTimeZone(0, "UTC")

        GregorianCalendar calendar = new GregorianCalendar(gmt)
        XMLGregorianCalendar xmlCalendar = new XMLGregorianCalendarImpl(calendar)

        audienceRestriction.getAudiences().add("service.example.com")
        conditions.getConditionsAndOneTimeUsesAndAudienceRestrictions().add(audienceRestriction)

        nameID.setValue("1234567")
        nameID.setFormat("urn:oasis:names:tc:SAML:2.0:nameid-format:transient")

        subject.setNameID(nameID)

        authnRequest.setSignature(signature)
        authnRequest.setSubject(subject)
        authnRequest.setConditions(conditions)

        authnRequest.setForceAuthn(false)
        authnRequest.setAssertionConsumerServiceURL("http://service.example.com/sso/saml.post")
        authnRequest.setAttributeConsumingServiceIndex(0)
        authnRequest.setProviderName("service")
        authnRequest.setID("abe567de6-122wert67")
        authnRequest.setVersion("2.0")
        authnRequest.setIssueInstant(xmlCalendar)

        return authnRequest
    }

    def createSchemaInvalidAuthnRequest() {
		AudienceRestriction audienceRestriction = new AudienceRestriction()
        Conditions conditions = new Conditions()
        NameIDType nameID = new NameIDType()
        Subject subject = new Subject()
        Signature signature = new Signature()
        AuthnRequest authnRequest = new AuthnRequest()

        SimpleTimeZone gmt = new SimpleTimeZone(0, "UTC")

        GregorianCalendar calendar = new GregorianCalendar(gmt)
        XMLGregorianCalendar xmlCalendar = new XMLGregorianCalendarImpl(calendar)

        audienceRestriction.getAudiences().add("service.example.com")
        conditions.getConditionsAndOneTimeUsesAndAudienceRestrictions().add(audienceRestriction)

        nameID.setValue("1234567")
        nameID.setFormat("urn:oasis:names:tc:SAML:2.0:nameid-format:transient")

        subject.setNameID(nameID)

        authnRequest.setSignature(signature)

        authnRequest.setForceAuthn(false)
        authnRequest.setAttributeConsumingServiceIndex(0)
        authnRequest.setProviderName("service")
        authnRequest.setVersion("2.0")
        authnRequest.setIssueInstant(xmlCalendar)

        return authnRequest
    }
}