package intient.xml.bind.impl

import groovy.xml.*
import groovy.mock.interceptor.MockFor

import org.oasis.saml2.assertion.*;
import org.oasis.saml2.protocol.*;
import org.w3._2000._09.xmldsig_.*;

import java.security.KeyStore
import java.security.PrivateKey
import java.security.cert.X509Certificate
import java.security.PublicKey

import javax.xml.datatype.XMLGregorianCalendar;
import com.sun.org.apache.xerces.internal.jaxp.datatype.XMLGregorianCalendarImpl;
import com.sun.xml.internal.bind.marshaller.NamespacePrefixMapper;

import org.w3c.dom.Document

import intient.xml.bind.impl.*
import intient.xml.constants.*
import intient.xml.keys.*
import intient.xml.namespace.*
import intient.xml.exception.*

class UnmarshallerImplTest extends GroovyTestCase {

	def keyAlias = 'testing'
	def packageName = "org.oasis.saml2.protocol"

	void testUnmarshalSignedByteArrayPK() {
		def unmarshaller = setupUnmarshaller()
		def xml = loadDocument()
		def crypto = loadCrypto()

        def authnRequest = unmarshaller.unmarshalSigned(crypto.pk, xml)
        assertTrue authnRequest.assertionConsumerServiceURL.equals("http://service.example.com/sso/saml.post")
        assertTrue authnRequest.signature.signedInfo.references.get(0).uri.equals("#abe567de6-122wert67")
        assertTrue authnRequest.subject.nameID.format.equals("urn:oasis:names:tc:SAML:2.0:nameid-format:transient")
	}

	void testUnmarshalSignedByteArrayResolver() {
		def unmarshaller = setupUnmarshallerWithResolver()
		def xml = loadDocument()

        def authnRequest = unmarshaller.unmarshalSigned(xml)
        assertTrue authnRequest.assertionConsumerServiceURL.equals("http://service.example.com/sso/saml.post")
        assertTrue authnRequest.signature.signedInfo.references.get(0).uri.equals("#abe567de6-122wert67")
        assertTrue authnRequest.subject.nameID.format.equals("urn:oasis:names:tc:SAML:2.0:nameid-format:transient")
	}

	void testUnmarshalSignedNullByteArray() {
		def unmarshaller = setupUnmarshallerWithResolver()

        def failure = shouldFail(IllegalArgumentException.class) {
			def authnRequest = unmarshaller.unmarshalSigned((byte[])null)
		}
		assertTrue failure.contains("Invalid byte array supplied to translate operation")
	}

	void testUnmarshalSignedByteArrayPKInvalidSig() {
		def unmarshaller = setupUnmarshaller()
		def xml = loadInvalidSigDocument()
		def crypto = loadCrypto()
        
        def failure = shouldFail(ReferenceValueException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(crypto.pk, xml)
		}
		assertTrue failure.contains("Signature validation failed with invalid supplied reference, the request document has been tampered with or the parent element to this signature block has no ID attribute")
	}

	void testUnmarshalSignedByteArrayPKInvalidSigValue() {
		def unmarshaller = setupUnmarshaller()
		def xml = loadInvalidSigValueDocument()
		def crypto = loadCrypto()
        
        def failure = shouldFail(SignatureValueException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(crypto.pk, xml)
		}
		assertTrue failure.contains("Signature validation failed with invalid signature for supplied public key")
	}

	void testUnmarshalSignedByteArrayResolverInvalidSig() {
		def unmarshaller = setupUnmarshallerWithResolver()
		def xml = loadInvalidSigDocument()
        
        def failure = shouldFail(ReferenceValueException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(xml)
		}
		assertTrue failure.contains("Signature validation failed with invalid supplied reference, the request document has been tampered with or the parent element to this signature block has no ID attribute")
	}

	void testUnmarshalSchemaViolationSignedByteArrayPK() {
		def unmarshaller = setupUnmarshallerWithResolver()
		def xml = loadInvalidSchemaDocument()
		def crypto = loadCrypto()
		
        shouldFail(UnmarshallerException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(crypto.pk, xml)
		}
	}

	void testUnmarshalSchemaViolationSignedByteArrayResolver() {
		def unmarshaller = setupUnmarshallerWithResolver()
		def xml = loadInvalidSchemaDocument()
		
        shouldFail(UnmarshallerException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(xml)
		}
	}

	void testUnmarshalSignedByteArrayResolverInvalidSigValue() {
		def unmarshaller = setupUnmarshallerWithResolver()
		def xml = loadInvalidSigValueDocument()
        
        def failure = shouldFail(SignatureValueException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(xml)
		}
		assertTrue failure.contains("Signature validation failed with invalid signature for supplied public key")
	}

	void testUnmarshalSignedByteArrayPKWrongKey() {
		def unmarshaller = setupUnmarshaller()
		def xml = loadDocument()
		def crypto = loadCrypto2()
        
        def failure = shouldFail(SignatureValueException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(crypto.pk, xml)
		}
		assertTrue failure.contains("Signature validation failed with invalid signature for supplied public key")
	}
	
	void testUnmarshalSignedByteArrayResolverNoKey() {
		def unmarshaller = setupUnmarshallerWithResolver2()
		def xml = loadDocument()
        
        def failure = shouldFail(KeySelectorException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(xml)
		}
		assertTrue failure.contains("Key with issuerName CN=Unknown, OU=Unknown, O=Unknown, L=Unknown, ST=Unknown, C=Unknown and serialNumber 1251768795 is unknown")
	}

	void testUnmarshalUnsignedByteArray() {
		def unmarshaller = setupUnmarshaller()
		def xml = loadDocument()

        def authnRequest = unmarshaller.unmarshalUnsigned(xml)
        assertTrue authnRequest.assertionConsumerServiceURL.equals("http://service.example.com/sso/saml.post")
        assertTrue authnRequest.signature.signedInfo.references.get(0).uri.equals("#abe567de6-122wert67")
        assertTrue authnRequest.subject.nameID.format.equals("urn:oasis:names:tc:SAML:2.0:nameid-format:transient")
	}

	void testUnmarshalUnsignedNullByteArray() {
		def unmarshaller = setupUnmarshallerWithResolver()

        def failure = shouldFail(IllegalArgumentException.class) {
			def authnRequest = unmarshaller.unmarshalUnsigned((byte[])null)
		}
		assertTrue failure.contains("Invalid byte array supplied to translate operation")
	}

	void testUnmarshalSchemaViolationUnsignedByteArray() {
		def unmarshaller = setupUnmarshallerWithResolver()
		def xml = loadInvalidSchemaDocument()
		
        shouldFail(UnmarshallerException.class) {
			def authnRequest = unmarshaller.unmarshalUnsigned(xml)
		}
	}

	void testUnmarshalSignedDocumentPK() {
		def unmarshaller = setupUnmarshaller()
		def crypto = loadCrypto()

        def authnRequest = unmarshaller.unmarshalSigned(crypto.pk, buildDocument())
        assertTrue authnRequest.assertionConsumerServiceURL.equals("http://service.example.com/sso/saml.post")
        assertTrue authnRequest.signature.signedInfo.references.get(0).uri.equals("#abe567de6-122wert67")
        assertTrue authnRequest.subject.nameID.format.equals("urn:oasis:names:tc:SAML:2.0:nameid-format:transient")
	}

	void testUnmarshalSignedDocumentPKInvalidSig() {
		def unmarshaller = setupUnmarshaller()
		def crypto = loadCrypto()
        
        def failure = shouldFail(ReferenceValueException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(crypto.pk, buildInvalidSigDocument())
		}
		assertTrue failure.contains("Signature validation failed with invalid supplied reference, the request document has been tampered with or the parent element to this signature block has no ID attribute")
	}

	void testUnmarshalSignedDocumentPKInvalidSigValue() {
		def unmarshaller = setupUnmarshaller()
		def crypto = loadCrypto()
        
        def failure = shouldFail(SignatureValueException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(crypto.pk, buildInvalidSigValueDocument())
		}
		assertTrue failure.contains("Signature validation failed with invalid signature for supplied public key")
	}

	void testUnmarshalNullSignedDocumentPK() {
		def unmarshaller = setupUnmarshallerWithResolver()
		def crypto = loadCrypto()
        def failure = shouldFail(IllegalArgumentException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(crypto.pk, (Document)null)
		}
		assertTrue failure.contains("Invalid document supplied to unmarshal operation")
	}

	void testUnmarshalSchemaViolationSignedDocumentPK() {
		def unmarshaller = setupUnmarshallerWithResolver()
		def crypto = loadCrypto()
        shouldFail(UnmarshallerException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(crypto.pk, buildInvalidSchemaDocument())
		}
	}

	void testUnmarshalSignedDocumentPKWrongKey() {
		def unmarshaller = setupUnmarshaller()
		def crypto = loadCrypto2()
        
        def failure = shouldFail(SignatureValueException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(crypto.pk, buildInvalidSigDocument())
		}
		assertTrue failure.contains("Signature validation failed with invalid signature for supplied public key")
	}

	void testUnmarshalSignedDocumentResolver() {
		def unmarshaller = setupUnmarshallerWithResolver()

        def authnRequest = unmarshaller.unmarshalSigned(buildDocument())
        assertTrue authnRequest.assertionConsumerServiceURL.equals("http://service.example.com/sso/saml.post")
        assertTrue authnRequest.signature.signedInfo.references.get(0).uri.equals("#abe567de6-122wert67")
        assertTrue authnRequest.subject.nameID.format.equals("urn:oasis:names:tc:SAML:2.0:nameid-format:transient")
	}

	void testUnmarshalSignedDocumentResolverInvalidSig() {
		def unmarshaller = setupUnmarshallerWithResolver()
        
        def failure = shouldFail(ReferenceValueException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(buildInvalidSigDocument())
		}
		assertTrue failure.contains("Signature validation failed with invalid supplied reference, the request document has been tampered with or the parent element to this signature block has no ID attribute")
	}

	void testUnmarshalSignedDocumentResolverInvalidSigValue() {
		def unmarshaller = setupUnmarshallerWithResolver()
        
        def failure = shouldFail(SignatureValueException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(buildInvalidSigValueDocument())
		}
		assertTrue failure.contains("Signature validation failed with invalid signature for supplied public key")
	}

	void testUnmarshalNullSignedDocumentResolver() {
		def unmarshaller = setupUnmarshallerWithResolver()

        def failure = shouldFail(IllegalArgumentException.class) {
			def authnRequest = unmarshaller.unmarshalSigned((Document)null)
		}
		assertTrue failure.contains("Invalid document supplied to unmarshal operation")
	}

	void testUnmarshalSchemaViolationSignedDocumentResolver() {
		def unmarshaller = setupUnmarshallerWithResolver()

        shouldFail(UnmarshallerException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(buildInvalidSchemaDocument())
		}
	}

	void testUnmarshalSignedDocumentResolverNoKey() {
		def unmarshaller = setupUnmarshallerWithResolver2()
        
        def failure = shouldFail(KeySelectorException.class) {
			def authnRequest = unmarshaller.unmarshalSigned(buildDocument())
		}
		assertTrue failure.contains("Key with issuerName CN=Unknown, OU=Unknown, O=Unknown, L=Unknown, ST=Unknown, C=Unknown and serialNumber 1251768795 is unknown")
	}

	void testUnmarshalUnsignedDocument() {
		def unmarshaller = setupUnmarshaller()

        def authnRequest = unmarshaller.unmarshalUnsigned(buildDocument())
        assertTrue authnRequest.assertionConsumerServiceURL.equals("http://service.example.com/sso/saml.post")
        assertTrue authnRequest.signature.signedInfo.references.get(0).uri.equals("#abe567de6-122wert67")
        assertTrue authnRequest.subject.nameID.format.equals("urn:oasis:names:tc:SAML:2.0:nameid-format:transient")
	}

	void testUnmarshalNullUnsignedDocument() {
		def unmarshaller = setupUnmarshallerWithResolver()

        def failure = shouldFail(IllegalArgumentException.class) {
			def authnRequest = unmarshaller.unmarshalUnsigned((Document)null)
		}
		assertTrue failure.contains("Invalid document supplied to unmarshal operation")
	}

	void testUnmarshalSchemaViolationUnsignedDocument() {
		def unmarshaller = setupUnmarshallerWithResolver()

        shouldFail(UnmarshallerException.class) {
			def authnRequest = unmarshaller.unmarshalUnsigned(buildInvalidSchemaDocument())
		}
	}

	// Helpers
    def setupUnmarshaller() {
    	def schemaList = loadSchema() 	
    	def unmarshaller = new UnmarshallerImpl<AuthnRequest>(packageName, schemaList, null)
    }

    def setupUnmarshallerWithResolver() {
    	def schemaList = loadSchema()
    	def crypto = loadCrypto()
    	
    	def selectKey = { keyName ->
    		if(keyName.equals(crypto.keyAlias))
    			return crypto.pk
    		throw new KeySelectorException("Key $keyName is unknown")
    	}

    	def extendedselectKey = { issuerName, serialNumber ->   	
			if(issuerName.equals(crypto.cert.issuerDN.name) && serialNumber == crypto.cert.serialNumber)
				return crypto.pk	
			throw new KeySelectorException("Key with issuerName $issuerName and serialNumber $serialNumber is unknown")
    	}

    	def externalKeyResolver = [selectKey:selectKey, selectKey:extendedselectKey] as PublicKeySelector	
    	def unmarshaller = new UnmarshallerImpl<AuthnRequest>(packageName, schemaList, externalKeyResolver)
    }

    def setupUnmarshallerWithResolver2() {
    	def schemaList = loadSchema()
    	def crypto = loadCrypto2()
    	
    	def selectKey = { keyName ->
    		if(keyName.equals(crypto.keyAlias))
    			return crypto.pk
    		throw new KeySelectorException("Key $keyName is unknown")
    	}

    	def extendedselectKey = { issuerName, serialNumber ->   	
			if(issuerName.equals(crypto.cert.issuerDN.name) && serialNumber == crypto.cert.serialNumber)
				return crypto.pk	
			throw new KeySelectorException("Key with issuerName $issuerName and serialNumber $serialNumber is unknown")
    	}

    	def externalKeyResolver = [selectKey:selectKey, selectKey:extendedselectKey] as PublicKeySelector	
    	def unmarshaller = new UnmarshallerImpl<AuthnRequest>(packageName, schemaList, externalKeyResolver)
    }

    def loadSchema() {
    	def assertion = new File("src/test/groovy/intient/xml/bind/saml-schema-assertion-2.0.xsd")
    	def metadata = new File("src/test/groovy/intient/xml/bind/saml-schema-metadata-2.0.xsd")
    	def protocol = new File("src/test/groovy/intient/xml/bind/saml-schema-protocol-2.0.xsd")

    	def schemas = [assertion.toURL(), protocol.toURL(), metadata.toURL()]
        return schemas
    }

    def loadCrypto() {
    	KeyStore ks = KeyStore.getInstance("JKS")
        FileInputStream fis = new FileInputStream("src/test/groovy/intient/xml/bind/keystore.ks")
        char[] passwd = ['t', 'e', 's', 't', 'i', 'n', 'g']
        ks.load(fis, passwd)

        PrivateKey priv = (PrivateKey) ks.getKey(keyAlias, passwd)
        X509Certificate cert = ks.getCertificate(keyAlias)
        PublicKey pk = cert.getPublicKey()

        def keyStore = [keyAlias:keyAlias, privateKey:priv, pk:pk, cert:cert]
    }

    def loadCrypto2() {
    	KeyStore ks = KeyStore.getInstance("JKS")
        FileInputStream fis = new FileInputStream("src/test/groovy/intient/xml/bind/keystore2.ks")
        char[] passwd = ['t', 'e', 's', 't', 'i', 'n', 'g']
        ks.load(fis, passwd)

        PrivateKey priv = (PrivateKey) ks.getKey(keyAlias, passwd)
        X509Certificate cert = ks.getCertificate(keyAlias)
        PublicKey pk = cert.getPublicKey()

        def keyStore = [keyAlias:keyAlias, privateKey:priv, pk:pk, cert:cert]
    }

    def loadDocument() {
    	def file = new File('testdata/authnrequest-valid.xml')
    	return file.readBytes()
    }

    def loadInvalidSigDocument() {
    	def file = new File('testdata/authnrequest-invalid-sig.xml')
    	return file.readBytes()
    }

    def loadInvalidSigValueDocument() {
    	def file = new File('testdata/authnrequest-invalid-sigvalue.xml')
    	return file.readBytes()
    }

    def loadInvalidSchemaDocument() {
    	def file = new File('testdata/authnrequest-invalid-schema.xml')
    	return file.readBytes()
    }

    def buildDocument() {
    	def domBuilder = DOMBuilder.newInstance()
		Document document = domBuilder.parse(new FileReader('testdata/authnrequest-valid.xml'))
    }

    def buildInvalidSigDocument() {
    	def domBuilder = DOMBuilder.newInstance()
		Document document = domBuilder.parse(new FileReader('testdata/authnrequest-invalid-sig.xml'))
    }

    def buildInvalidSigValueDocument() {
    	def domBuilder = DOMBuilder.newInstance()
		Document document = domBuilder.parse(new FileReader('testdata/authnrequest-invalid-sigvalue.xml'))
    }

    def buildInvalidSchemaDocument() {
    	def domBuilder = DOMBuilder.newInstance()
		Document document = domBuilder.parse(new FileReader('testdata/authnrequest-invalid-schema.xml'))
    }
}